#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include <fcntl.h>

#define SERVER_PGM "usrpgmsv3052f.bin"
#define START_ADDR 0xffdf10

#define BAUDRATE B19200

#define SEQSIZE 128
#define SENDSIZE 0x2000
#define RAMSIZE 0x2000

#define NBLKS 16


#ifndef ROMWRITE
 #warning Rom pgm is disabled
#endif

const long blocks[NBLKS] = 
  {
    0x0, 0x1000, 0x2000, 0x3000, 0x4000, 0x5000, 0x6000, 0x7000,
    0x8000,
    0x10000, 0x20000, 0x30000, 0x40000, 0x50000, 0x60000, 0x70000
  };
const long sizes[NBLKS] = 
  {
    0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000,
    0x8000,
    0x10000, 0x10000, 0x10000, 0x10000, 0x10000, 0x10000, 0x10000,
  };
#define MAXBLKSIZE 0x10000
#define ROMSIZE 0x80000
char iserased[NBLKS];
char iswritten[ROMSIZE / SEQSIZE];

int ttyfd;

int SendCMD(const char *fmt, ...);

int bSend (unsigned char data, unsigned int address, int isflush);
int bPGM (unsigned char data, unsigned int address, int isflush);

int SetupPGM(void);
int FinishPGM(void);
int AwaitACK(void);
int ReadMOT(FILE *fp, int (*bwrite)(unsigned char data, unsigned int address, int isflush));
int RunBIN(FILE *fp, unsigned int start);
int PGMBLK(char *buf, int blk);


int fdputc(int tty, unsigned char c) 
{
  return write(tty, &c, 1);
}
int fdput16(int tty, unsigned short i)
{
  fdputc(tty, i >> 8); //MSB first
  fdputc(tty, i % 0x100);
  return 0;
}
int fdput24(int tty, unsigned int i)
{
  fdput16(tty, i >> 8); //MSB first
  fdputc(tty, i % 0x100);
  return 0;
}

unsigned short MakeCheckSum(unsigned char *buf, int cnt);

int ROMPGM(FILE *srcfp)
{
  unsigned char c;
  int ret;
  FILE *pfp;

  pfp = fopen(SERVER_PGM, "r");
  if(!pfp)
    {
      printf("Opening PGM binary failed\n");
      return -1;
    }
  printf("PGM binary transferring...\n");
  if((ret = RunBIN(pfp, START_ADDR)))
    return ret;

  fdputc(ttyfd, 'c');
  fdput16(ttyfd, 20000);
  read(ttyfd, &c, 1);
  if(c != 20000 / 0x100)
    {
      printf("Cannot establish connection\n");
      return -1;
    }
  printf("PGM binary started\n");
  fclose(pfp);
      
  printf("Starting PGM...\n");
  if((ret = SetupPGM()))
    return ret;

  printf("Writing...\n");
  if((ret = ReadMOT(srcfp, bPGM)))
    return ret;
  return 0;
}

int main(int argc, char **argv) 
{
  FILE *srcfp;
  struct termios ttyios;
  int ret;

  if((ttyfd = open(argv[3], O_RDWR)) == -1)
    {
      fprintf(stderr, "tty open failed\n");
      return -1;
    }
  
  tcsetpgrp(ttyfd, getpgrp());
  tcgetattr(ttyfd, &ttyios);
  cfsetispeed(&ttyios, BAUDRATE);
  cfsetospeed(&ttyios, BAUDRATE);
  cfmakeraw(&ttyios);
  ttyios.c_cflag |= HUPCL | CLOCAL ;
  ttyios.c_cflag &= ~( PARENB | CSTOPB );
  if(tcsetattr(ttyfd, TCSANOW, &ttyios ) == -1){
    fprintf(stderr, "stty error\n" );
    return -1;
  }
  tcflush(ttyfd, TCIOFLUSH);

  SendCMD("\n");
//  AwaitACK();
  usleep(100000);
  tcflush(ttyfd, TCIOFLUSH);
  
  if(!strcmp(argv[1], "-rom"))
    {
      srcfp = fopen(argv[2], "r");
      if(!srcfp)
	{
	  printf("Opening user PGM failed\n");
	  return -1;
	}
      if((ret = ROMPGM(srcfp)))
	{
	  printf("Failed %d\n", ret);
	}
      fclose(srcfp);
      printf("Finishing PGM...\n");
      if((ret = FinishPGM()))
	{
	  printf("Failed %d\n", ret);
	  return -1;
	}
    }
  if(!strcmp(argv[1], "-ram"))
    {
      srcfp = fopen(argv[2], "r");
      if(!srcfp)
	{
	  printf("Opening user PGM failed\n");
	  return -1;
	} 
      printf("Writing...\n");
      if(ReadMOT(srcfp, bSend))
	{
	  printf("Failed\n");
	  return -1;
	}
      fclose(srcfp);
    }
  close(ttyfd);
  return 0;
}

int RunBIN(FILE *fp, unsigned int start)
{
  unsigned int size = 0;
  char buf[RAMSIZE];
  char *pbuf = buf;
  unsigned short sum;

  while(!feof(fp) && (size < RAMSIZE))
    {
      *pbuf++ = fgetc(fp);
      size++;
    }

  sum = MakeCheckSum(buf, size);
  SendCMD("load %X %X %X 1\n", start, size, (unsigned int)sum);
  write(ttyfd, ">", 1);
  write(ttyfd, buf, size);
  usleep(1000000);
  tcflush(ttyfd, TCIOFLUSH);
//  AwaitACK();
  return 0;
}

int SetupPGM(void)
{
  unsigned char c;
  int i;

  for(i = 0; i < ROMSIZE / SEQSIZE; i++)
    iswritten[i] = 0;
  for(i = 0; i < NBLKS; i++)
    iserased[i] = 0;

  read(ttyfd, &c, 1);
  if(c != ':') return c;
  fdputc(ttyfd, 's');
  read(ttyfd, &c , 1);
  if(c != 0) return c;
  return 0;
}

int FinishPGM(void)
{
  unsigned char c;

  read(ttyfd, &c, 1);
  if(c != ':') return c;
  fdputc(ttyfd, 'f');
  read(ttyfd, &c , 1);
  if(c != 0) return c;

  printf("Target is going to halt.\n");

  read(ttyfd, &c, 1);
  if(c != ':') return c;
  fdputc(ttyfd, 'h');

  printf("PGM succeed\n");
  return 0;
}

unsigned short MakeCheckSum(unsigned char *buf, int cnt)
{
  unsigned short sum = 0;
  for(;cnt > 0; cnt--)
    {
      sum += *buf++;
    }
  return sum;
}

int SendCMD(const char *fmt, ...)
{
  va_list ap;
  char buf[256];
  char c;

  va_start(ap, fmt);
  vsnprintf(buf, 256, fmt, ap);
  dprintf(ttyfd, "%s", buf);

  while(read(ttyfd, &c, 1), c != '\n') putchar(c);
  putchar(c);
  va_end(ap);
  return 0;
}

int AwaitACK(void)
{
  char c;

  while(read(ttyfd, &c, 1), c != ':') putchar(c);
  putchar(c);
  return 0;
}

int BlkSend(char *buf, unsigned int start, int cnt)
{
  unsigned short sum;

  sum = MakeCheckSum(buf, cnt);
  SendCMD("load %X %X %X\n", (unsigned int)start, (unsigned int)cnt, (unsigned int)sum);
  write(ttyfd, ">", 1);
  write(ttyfd, buf, cnt);
  usleep(1000000);
  tcflush(ttyfd, TCIOFLUSH);
//  AwaitACK();
  return 0;
}

int ReadBLK(char *buf, int blk)
{
  unsigned char c;
  int addr, i;

  for(addr = blocks[blk]; addr < blocks[blk] + sizes[blk]; addr += SEQSIZE)
    {
      read(ttyfd, &c, 1);
      if(c != ':') return c;
      
      fdputc(ttyfd, 'r');
      fdput24(ttyfd, addr);
      fdput16(ttyfd, SEQSIZE);
      
      read(ttyfd, &c, 1);
      
      for(i = 0; i < SEQSIZE; i++) read(ttyfd, buf++, 1);

      i = MakeCheckSum(buf - SEQSIZE, SEQSIZE);
      if(i % 0x100 != c)
	{
	  printf("Read error %x %x\n", i, c);
	}

      read(ttyfd, &c, 1);

      //      buf += SEQSIZE;
    }
  return 0;
}

int FindBLK(unsigned int address)
{
  int i;

  for(i = 0; i < NBLKS; i++)
    {
      if(blocks[i] + sizes[i] > address) return i;
    }
  return -1;
}

int EraseBlk(int block)
{
  unsigned char c;

#ifdef ROMWRITE  
  read(ttyfd, &c, 1);
  if(c != ':') return c;
  fdputc(ttyfd, 'e');

  c = block;  write(ttyfd, &c, 1);

  read(ttyfd, &c, 1);
  if(c != 0) return c;
  iserased[block] = 1;
#endif
  return 0;
}

int PGMBLK(char *buf, int blk)
{
  unsigned char c;
  int ret;
  unsigned int col;

  if(!iserased[blk])
    {
      printf("Erase block: #%d\n", blk);
      if((ret = EraseBlk(blk)))
	{
	  printf("failed #%d\n", ret);
	  return ret;
	}
      iserased[blk] = 1;
    }
  for(col = blocks[blk] / SEQSIZE; col < (blocks[blk] + sizes[blk]) / SEQSIZE; col++)
    {
      if(iswritten[col])
	{
	  printf("Cannnot overwrite ADDR: %6X\n", col * SEQSIZE);
	  return -1;
	}
      iswritten[col] = 1;
      
      printf("ADDR: %6X SUM: %4X\n", col * SEQSIZE, (unsigned int)MakeCheckSum(buf, SEQSIZE));
#ifdef ROMWRITE
      read(ttyfd, &c, 1);
      if(c != ':') return c;
      fdputc(ttyfd, 'p');
      fdputc(ttyfd, blk);
      fdput24(ttyfd, col * SEQSIZE);
      fdput16(ttyfd, MakeCheckSum(buf, SEQSIZE));
      
      write(ttyfd, buf, SEQSIZE);
      
      read(ttyfd, &c, 1);
      
      if(c != 0)
	{
	  printf("PGM failed %d\n", c);
	  return c;
	}
#endif
      buf += SEQSIZE;
    }
  return 0;
}

int bPGM (unsigned char data, unsigned int address, int isflush)
{
  static unsigned char buf[MAXBLKSIZE];
  static int cnt = 0;
  static int blk;
  static int ischanged;
  
  if(cnt && !isflush && (FindBLK(address) == blk))
    {
      if(buf[address - blocks[blk]] != data)
	{
	  ischanged = 1;
	  buf[address - blocks[blk]] = data;
	}
      cnt++;
    }
  else
    {
      if(cnt)
	{
	  printf("block: %d counts: %d\n", blk, cnt);

	  if(ischanged)
	    {
	      if(PGMBLK(buf, blk)) return -1;
	    }
	  else
	    {
	      printf("Skipping block: #%d\n", blk);
	    }
	}
      cnt = 0;
      if(!isflush)
	{
	  blk = FindBLK(address);
	  if(blk == -1)
	    {
	      printf("Cannot find block for %6X\n", address); 
	      return 0;
	    }
	  ischanged = 0;
	  ReadBLK(buf, blk);
	  if(buf[address - blocks[blk]] != data)
	    {
	      ischanged = 1;
	      buf[address - blocks[blk]] = data;
	    }
	  cnt = 1;
	}
    }
  return 0;
}

int bSend (unsigned char data, unsigned int address, int isflush)
{
  static unsigned char buf[SENDSIZE];
  static int cnt = 0;
  static unsigned int start = 0xffffff;

  if((start + cnt == address) && cnt && !isflush)
    {
      buf[cnt++] = data;
      if(cnt < SENDSIZE) return 0;
      BlkSend(buf, start, cnt);
      cnt = 0;
    }
  else
    {
      if(cnt) BlkSend(buf, start, cnt);
      cnt = 0;
      if(!isflush)
	{
	  buf[0] = data;
	  start = address;
	  cnt = 1;
	}
    }
  return 0;
}

int ReadMOT(FILE *fp, int (*bwrite)(unsigned char data, unsigned int address, int isflush))
{
  char cbuf[256];
  unsigned int length, address, byte;
  int n;
  int ret;

  while(fgets(cbuf, 256, fp))
    {
      if(cbuf[0] != 'S')
	{
	  printf("Bad S Format\n");
	  return -1;
	}
      switch(cbuf[1])
	{
	case '0':
	  break;
	case '1':
 	  if(sscanf(cbuf, "S1%2x%4x", &length, &address) != 2)
	    {
	      printf("Bad S Format\n");
	      return -1;
	    }
	  length -= 3;
	  for(n = 0; n < length; n++)
	    {
	      if(sscanf(cbuf + 2*n + 8, "%2x", &byte) != 1)
		{
		  printf("Bad S Format\n");
		  return -1;
		}
	      if((ret = bwrite(byte, address++, 0)))
		{
		  printf("Writing Failed\n");
		  return ret;
		}
	    }
	  break;
	case '2':
 	  if(sscanf(cbuf, "S2%2x%6x", &length, &address) != 2)
	    {
	      printf("Bad S Format\n");
	      return -1;
	    }
	  length -= 4;
	  for(n = 0; n < length; n++)
	    {
	      if(sscanf(cbuf + 2*n + 10, "%2x", &byte) != 1)
		{
		  printf("Bad S Format\n");
		  return -1;
		}
	      if((ret = bwrite(byte, address++, 0)))
		{
		  printf("Writing Failed\n");
		  return ret;
		}
	    }
	  break;
	case '8':
	case '9':
	  printf("Flushing buffer\n");
	  if((ret = bwrite(byte, 0, 1)))
	    {
	      printf("Writing Failed\n");
	      return ret;
	    }
	  return 0;
	default:	
	  printf("Bad S Format\n");
	  return -1;
	}
    }
  printf("Bad S Format\n");
  return -1;
}

